//Below is the 78 chars line!
//----------------------------------------------------------------------------
//Jason Winnebeck

#include <string.h>
#include <stdio.h>
#include <errno.h>

#include <sys/socket.h> //consts and procedures
#include <sys/types.h>  //for timeval
//#include <sys/types.h>
#define LinkListConflict
#include <unistd.h>     //for close()
#include <sys/time.h>   //for user select
#include <arpa/inet.h>  //for sockadd_in !!!
#include <fcntl.h>      //for fcntl

#include <stdlib.h>     //for atexit()

#include "server.h"
#include "window.h"
#include <curses.h>

int main() {
  atexit(DisplayError);
  initscr();
  noecho();
  InitColors();
  if (Init() == ERROR)
    return ERROR;

  //Setup the TCP/IP listening socket
  int lsock = SetupListen();
  if (lsock == ERROR) {
    Message << "Listen socket error!\n";
    return ERROR;
  }
  Message << "Listening. . .\n";
  if (SetupBroadcast() == ERROR) {
    Message << "Broadcasting socket error!\n";
    return ERROR;
  }

  bool exit = false;

  while (!exit) {
    fd_set readfds;

    FD_ZERO(&readfds);
    FD_SET(bsock, &readfds);
    FD_SET(lsock, &readfds);
    FD_SET(STDIN, &readfds);

    //Do a "max" operation
    int maxsock;
    if (bsock < lsock)
      maxsock = lsock + 1;
    else
      maxsock = bsock + 1;

    if (select(maxsock, &readfds, NULL, NULL, NULL) == ERROR)
      return ERROR;

    if (FD_ISSET(lsock, &readfds)) {
      Message << "Sensed something\n";
      if (NewCli(lsock) == ERROR)
        Message << "Client join error!!!\n";
    }

    if (FD_ISSET(bsock, &readfds)) {
      Clidata* cli;   //Check for broadcasting messages
      RecvFrom(cli);

      if (cli) {      //If it didn't block/have an error
        switch(buf[0]) {
          case MSG:
            SendMsg(cli, &buf[1]);
            break;
          case EXIT:
            LeaveCli(cli);
            break;
          case PING:
            break;
        }
      }
    }

    if (FD_ISSET(STDIN, &readfds)) {  //Check for user exit
      char ch = (char)getch();
      if (ch == '\n')
        exit = true;
        //WE NEED TO SEND CLIENTS SERVER EXIT NOTICE!!!
    }

    //    __dpmi_yield(); //Don't chew up all the resources
  }

  //Clean-up
  close(lsock);
  close(bsock);

  endwin();
  printf("Server exited successfully without error.\n");
  return 0;
}

void SendMsg(Clidata* cli, char* msg) {
  int newstrlen =
    strlen(msg) +       //The size of the actual message
    strlen(cli->name) + //Plus the size of the person's name
    5;                  //Header, the <>, the space, NULL

  char* temp = new char[newstrlen+50];
  sprintf(temp, "%c<%s> %s", MSG, cli->name, msg);

  Broadcast(temp);
  delete[] temp;
}

void SendOut(Clidata* cli, char* msg) {
  int msglen = strlen(msg) + 1; //We DO want to send the NULL character
  int sent = 0;    //Amount of data sent so far
  while (sent < msglen) {
    int result;
    result = sendto(bsock, &msg[sent], msglen-sent, (unsigned int)NULL, (sockaddr*)(&cli->loc), sizeof(sockaddr));
    if (result == ERROR)
      exit(ERROR);
    else
      sent += result;
  }
  gout += sent;  //Update connection stats
  UpdateStats();
}

void RecvFrom(Clidata* &cli) {
  sockaddr_in CliAddr;
  int size = sizeof(CliAddr);
  int result = recvfrom(bsock, buf, BUFLEN, (unsigned int)NULL, (sockaddr*)&CliAddr, (size_t*)&size);
  if (result != ERROR) {
    //Using loop method for more security and invalid ptr access protection
    cli = NULL; //No client if this gets through
    int l = Clients.Length();
    Clients.SetToStart();
    for (int c=0; c<l; c++) {
      sockaddr_in t = Clients._GetNext()->loc; //Reduce if clutter
      if ((t.sin_addr.s_addr == CliAddr.sin_addr.s_addr) && (t.sin_port == CliAddr.sin_port))
        cli = Clients._GetNext();
      Clients._Advance();
    }
    gin += result;
    UpdateStats();
  } else {
    cli = NULL; //No client
    Message << "Client recieve error!\n";
  }
}

void Broadcast(char* msg) {
  Clients.SetToStart();
  int len = Clients.Length();
  for (int c=0; c<len; c++) {
    SendOut(Clients._GetNext(), msg);
    Clients._Advance();
  }
  Message << msg << '\n'; //Let console see broadcasts
}

void BroadcastPing() {
}

int SetupListen() {
  sockaddr_in ina;
  bzero(&ina, sizeof(ina));
  ina.sin_family = AF_INET;
  ina.sin_port = htons(CPORT);
  ina.sin_addr.s_addr = htonl(INADDR_ANY);

  int list = socket(AF_INET, SOCK_STREAM, 0);
  if (list == ERROR)
    return ERROR;

  if (bind(list, (sockaddr *)&ina, sizeof(sockaddr)) == ERROR)
    return ERROR;

  if (listen(list, MAXCLI) == ERROR)
    return ERROR;

  return list;
}

int SetupBroadcast() {
  sockaddr_in ina;
  bzero(&ina, sizeof(ina));
  ina.sin_family = AF_INET;
  ina.sin_port = htons(BPORT);
  ina.sin_addr.s_addr = htonl(INADDR_ANY);

  bsock = socket(AF_INET, SOCK_DGRAM, 0);
  if (bsock == ERROR)
    return ERROR;

  if (bind(bsock, (sockaddr *)&ina, sizeof(sockaddr)) == ERROR)
    return ERROR;

  return 0;
}

void SendNewCli(Clidata* cli) {
  int newstrlen =
    2 +                //Packet header, NULL
    strlen(cli->name); //packet data

  char *msg = new char[newstrlen];
  sprintf(msg, "%c%s", char(NEW), cli->name);
  Message << "Broadcasting join message\n";
  Broadcast(msg);      //Send our packet.  Bye-bye packet!
  delete[] msg;
}

void LeaveCli(Clidata* cli) {
  int newstrlen =
    2 +                //Packet header and NULL
    strlen(cli->name); //packet data

  char *msg = new char[newstrlen];
  sprintf(msg, "%c%s", char(EXIT), cli->name);
  Broadcast(msg);      //Send our packet.  Bye-bye packet!
  delete[] msg;

  //Send verbal notice
  sprintf(buf, "* %s has left the chat!", cli->name);
  Broadcast(buf);

  //Delete this client from the list.  Delete by POINTER rather than name
  // in case of duplicate connections (if someone timed out?)
  int l = Clients.Length();
  Clients.SetToStart();
  for (int c=0; c<l; c++) {
    if (Clients._GetNext() == cli)
      Clients._DeleteThis();      //Yes but it SHOULDN'T delete more than 1!
    else
      Clients._Advance();
  }
  UpdateClilist();
}

void SendPing() {
  int newstrlen =
    2;                 //Packet header and NULL

  char *msg = new char[newstrlen];
  sprintf(msg, "%c", char(PING));
  Broadcast(msg);
  delete[] msg;
}

